/*
 * Amanda, The Advanced Maryland Automatic Network Disk Archiver
 * Copyright (c) 1991,1993 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: James da Silva, Systems Design and Analysis Group
 *			   Computer Science Department
 *			   University of Maryland at College Park
 */
#include "amanda.h"
#include "conffile.h"
#include "statfs.h"
#include "tapefile.h"
#include "tapeio.h"

char *pname = "amcheck";

int mailout, overwrite;
int disklow, tapebad;
char mailfile[256];

usage()
{
    error("Usage: amcheck [-m] [-w] <conf>");
}

int main(argc, argv)
int argc;
char **argv;
{
    char confdir[256], cmd[1024], label[80], datestamp[80];
    char *errstr, *confname, *tapename, *labelstr, *diskdir;
    generic_fs_stats_t fs;
    unsigned long disksize;
    tape_t *tp;


    erroutput_type = ERR_INTERACTIVE;

    argc--,argv++;	/* eat argv[0] */

    mailout = overwrite = 0;
    disklow = tapebad = 0;

    while(argc && argv[0][0] == '-') {
	if(!strcmp(*argv,"-m")) mailout = 1;
	else if(!strcmp(*argv,"-w")) overwrite = 1;
	else usage();
	argc--,argv++;
    }

    /* read in amanda configuration */

    if(argc != 1) usage();

    confname = *argv;

    sprintf(confdir, "%s/%s", CONFIG_DIR, confname);
    if(chdir(confdir) != 0)
        error("could not cd to confdir %s: %s", confdir, strerror(errno));

    if(read_conffile(CONFFILE_NAME))
        error("could not read amanda config file");

    if(read_tapelist(getconf_str(CNF_TAPELIST)))
        error("parse error in %s", getconf_str(CNF_TAPELIST));

    /* set up temp file if going to mail on error */

    if(mailout) {
	sprintf(mailfile, "/tmp/amcheck.%d", getpid());
	if(freopen(mailfile, "w", stdout) == NULL)
	    error("%s: %s", mailfile, strerror(errno));
    }

    /* check available disk space */

    diskdir = getconf_str(CNF_DISKDIR);
    disksize = getconf_int(CNF_DISKSIZE);
    if(get_fs_stats(diskdir, &fs) == -1)
	error("statfs %s: %s", diskdir, strerror(errno));

    if(fs.avail == -1) {
	printf("WARNING: cannot determine avail disk space.  %d KB requested.",
	       disksize);
    }
    else if(fs.avail < disksize) {
	printf("WARNING: disk space low: %d KB avail < %d KB requested.",
	       fs.avail, disksize);
	printf("\n         (please clear out cruft from %s's partition)\n",
	       diskdir);
	disklow = 1;
    }
    else printf("%d KB disk space available: that's plenty.\n", fs.avail);

    /* check that the tape is a valid amanda tape */

    tapename = getconf_str(CNF_TAPEDEV);
    labelstr = getconf_str(CNF_LABELSTR);

    if((errstr = tape_rdheader(tapename, datestamp, label)) != NULL) {
	printf("ERROR: %s: %s.\n", tapename, errstr);
	tapebad = 1;
    } else {

	tp = lookup_tapelabel(label);
	if(tp != NULL && tp->position < getconf_int(CNF_TAPECYCLE)) {
	    printf("ERROR: cannot overwrite active tape %s.\n", label);
	    tapebad = 1;
	}
	if(!match(labelstr, label)) {
	    printf("ERROR: label %s doesn't match labelstr \"%s\".\n",
		   label, labelstr);
	    tapebad = 1;
	}

    }
    if(tapebad) {
	tape_t *exptape = lookup_tapepos(getconf_int(CNF_TAPECYCLE));
	printf("       (expecting ");
	if(exptape != NULL) printf("tape %s or ", exptape->label);
	printf("a new tape)\n");
    }

    if(!tapebad && overwrite) {
	if((errstr = tape_writeable(tapename)) != NULL) {
	    printf("ERROR: tape %s label ok, but is not writeable.\n", label);
	    tapebad = 1;
	}
	else printf("Tape %s is writeable.\n", label);
    }
    else printf("NOTE: skipping tape-writeable test.\n");

    if(!tapebad)
	printf("Tape %s label ok.\n", label);

    if((tapebad || disklow) && mailout) {
	sprintf(cmd, 
	   "%s -s \"%s AMANDA PROBLEM: FIX BEFORE RUN, IF POSSIBLE\" %s < %s", 
		MAILER, getconf_str(CNF_ORG), getconf_str(CNF_MAILTO), 
		mailfile);
	fflush(stdout);
	if(system(cmd) != 0)
	    error("mail command failed: %s", cmd);
    }
    if(mailout)
	unlink(mailfile);
    exit(tapebad || disklow);
}


int match(regex, str)
char *regex, *str;
{
    char *result, *re_comp();

    if((result = re_comp(regex)) != NULL)
        error("bad regular expression \"%s\": %s", regex, result);
    return re_exec(str);
}
